import numpy as np
import pandas as pd
from tqdm import tqdm

class Underlying:
    """
    Object representing underlying asset of the option

    Parameters
    ----------
    :param spot_price: initial value of the asset
    :type spot_price: float
    :param r: risk-free rate
    :type r: float


    Examples
    --------
    Creation of an asset
    >>> asset = Underlying(spot_price=100, r=0.06)
    """
    def __init__(self, spot_price, r):
        self.spot_price = spot_price
        self.r = r
        self.GBM_calibrated = False
        self.Bootstrap_calibrated = False
        self.RT_GBM_calibrated = False
        self.JD_calibrated = False
        self.ssp_calibrated = False

    def calibrate_GBM(self, sigma, values_per_year_GBM=255):
        """
        Method needs to be called when one wants to simulate with GBM

        Parameters
        ----------
        :param sigma: volatility of an asset
        :type sigma: float
        :param values_per_year_GBM: how many time-steps we want to include, default: 255 (number of trading days per year)
        :type values_per_year_GBM: int

        Examples
        --------
        Creation of an asset
        >>> asset = Underlying(spot_price=100, r=0.06)
        >>> asset.calibrate_GBM(0.4, 255)
        """
        self.sigma = sigma
        self.values_per_year_GBM = values_per_year_GBM
        self.GBM_calibrated = True
        self.ssp_calibrated=False


    def calibrate_Bootstrap(self, Historical_Data, values_per_year_Bootstrap=255):
        """
        Method needs to be called when one wants to simulate with Bootstrap

        Parameters
        ----------
        :param Historical_Data: Historical values that will be used for bootstrapping
        :type Historical_Data: pandas DataFrame
        :param values_per_year_GBM: how many time-steps we want to include, default 255(number of trading days per year)
        :type values_per_year_GBM: int

        Examples
        --------
        Creation of an asset using historical data
        >>> hist_data = pd.read_csv("your_data.csv") #hist_data is one-row dataframe of past asset prices
        >>> asset = Underlying(100, 0.06)
        >>> asset.calibrate_Bootstrap(Historical_Data = hist_data, values_per_year_Bootstrap = 255)

        """

        self.Historical_Data = Historical_Data
        self.values_per_year_Bootstrap = values_per_year_Bootstrap
        self.Bootstrap_calibrated = True
        self.ssp_calibrated=False


    def calibrate_RT_GBM(self, sigma, time_steps, successors):
        """
        Method needs to be called when one wants to simulate with random tree

        Parameters
        ----------
        :param sigma: volatility of an asset
        :type sigma: float
        :param time_steps: how many time-steps we want to include, REMARK: time_steps > 5 will take ages to compute
        :type time_steps: int
        :param successors: how many new nodes in one step we create from one node,\n
         the greater the value, the narrower thi interval. REMARK: successors > 6 will take long
        :type successors: int

        Examples
        --------
        Creation of an asset
        >>> asset = Underlying(spot_price=100, r=0.06)
        >>> asset.calibrate_RT_GBM(0.4, 4, 3)
        """

        self.sigma = sigma
        self.time_steps = time_steps
        self.successors = successors
        self.RT_GBM_calibrated = True
        self.ssp_calibrated=False

    def calibrate_JD(self, sigma, jump_intensity, values_per_year_JD=255, a=0.2, b=0.25):
        """
        Method needs to be called when one wants to simulate with jump-diffusion model \n

        Parameters a and b are chosen so that ln(Yi) ~ N(a, b^2),
        with Yi being random variable describing multiplicative jumps ($Yi = S_{\tau_i}/S_{\tau_i-}$)

        Parameters
        ----------
        :param sigma: volatility of an asset
        :type sigma: float
        :param jump_intensity: parameter of Poisson process that is responsible for jumps in the model
        :type jump_intensity: float
        :param values_per_year_JD: how many time-steps we want to include, default 255(number of trading days per year)
        :type values_per_year_JD: int
        :param a:
        :type a: float
        :param b:
        :type b: float

        Examples
        --------
        Creation of an asset
        >>> asset = Underlying(spot_price=100, r=0.06)
        >>> asset.calibrate_JD(0.4, 0.4)
        """

        self.sigma = sigma
        self.jump_intensity = jump_intensity
        self.values_per_year_JD = values_per_year_JD
        self.a = a
        self.b = b
        self.JD_calibrated = True
        self.ssp_calibrated=False

    def simulate(self, mode, size, T=1):
        """
        Method that simulates asset development paths used in pricing

        Parameters
        ----------
        :param mode: Model in which simulations are performed, available values: "GBM", "Bootstrap", "RT_GBM", "JD"
        :type mode: str
        :param size: How many paths to simulate
        :type size: int
        :param T: Time horizon for simulation
        :type T: float

        Returns
        -------
        :return: Data frame having one simulation in each row
        :rtype: pandas.DataFrame

        Examples
        --------
        Creation of an asset
        >>> asset = Underlying(spot_price=100, r=0.06)
        >>> asset.calibrate_GBM(0.4, 255)
        >>> asset.simulate('GBM', 3, T=1)
                  0           1           2    ...         253         254         255
        0   99.999996  101.323642  100.757084  ...  155.495649  158.356571  154.573858
        1   99.999997  100.407545  101.755996  ...  121.491351  118.305042  121.353466
        2  100.000003  100.581243   99.541965  ...   92.986666   90.186351   90.603168
        [3 rows x 256 columns]

        >>> asset.calibrate_Bootstrap(hist_data, 255)
        >>> asset.simulate('Bootstrap', 3, T=1)

            0	        1	        2		...		254	        255
        0	100.0	98.388416	98.466800	...	174.518918	180.565038
        1	100.0	101.641189	101.824047  ...	89.923103	90.955017
        2	100.0	100.925334	101.018876	...	115.799699	114.339994
        [3 rows × 256 columns]
        """
        if mode == "GBM":
            if self.GBM_calibrated: #simulating in risk-free measure
                price_moments = np.arange(0, self.values_per_year_GBM * T + 1)
                Sigma = 1/self.values_per_year_GBM*np.minimum(np.tile(price_moments,(len(price_moments),1)),np.tile(price_moments.reshape(-1,1),(1,len(price_moments))))
                B = pd.DataFrame(np.random.multivariate_normal(size=size, mean= np.zeros(len(price_moments)), cov = Sigma))
                sims = self.spot_price*np.exp((self.r - 0.5 * self.sigma**2) * price_moments / self.values_per_year_GBM + self.sigma * B)
                #print(sims.shape)
                return(sims)
            else:
                raise Exception("GBM model has not been calibrated yet. Calibrate it with calibrate_GBM method first")
        elif mode == "Bootstrap":
            if self.Bootstrap_calibrated:
                price_moments = np.arange(0, round(self.values_per_year_Bootstrap * T) + 1)
                daily_mult_price_changes = (self.Historical_Data.iloc[:, (-round(self.values_per_year_Bootstrap * T)):].values / self.Historical_Data.iloc[:, (-round(self.values_per_year_Bootstrap * T)-1):(-1)].values)[0]
                sims = pd.DataFrame(np.reshape(np.random.choice(daily_mult_price_changes,size=len(price_moments)*size,replace=True),newshape=(size,round(self.values_per_year_Bootstrap * T) + 1)))
                sims.iloc[:, 0] = 1
                return(self.spot_price*sims.cumprod(axis="columns"))
            else:
                raise Exception("Boostrap model has not been calibrated yet. Calibrate it with calibrate_Bootstrap method first")
        elif mode == "RT_GBM":
            if self.RT_GBM_calibrated:
                delta_t = T/self.time_steps
                sims = []
                for _ in tqdm(range(size)):
                    sim = pd.DataFrame(np.zeros((self.successors ** self.time_steps, self.time_steps+1)))
                    sim.iloc[0, 0] = self.spot_price
                    for i in range(self.time_steps):
                        for j in range(self.successors ** i):
                            norm_dist = np.random.normal(loc=0, scale=1, size=self.successors)
                            vec = sim.iloc[j,i] * np.exp((self.r - 0.5 * self.sigma **2) * delta_t + self.sigma * np.sqrt(delta_t) * norm_dist)
                            sim.iloc[(j*self.successors):((j+1)*self.successors), i+1] = vec
                    sims.append(sim)
                return sims
            else:
                raise Exception("Random tree model has not been calibrated yet. Calibrate it with calibrate_RT_GBM method first")
        elif mode == "JD":
            if self.JD_calibrated:
                dt = T / self.values_per_year_JD
                ti = int(self.values_per_year_JD * T)
                l = self.jump_intensity
                simulated_paths = np.zeros([size, ti + 1])

                simulated_paths[:, 0] = self.spot_price

                Z_1 = np.random.normal(size=[size, ti])
                Z_2 = np.random.normal(size=[size, ti])
                Poisson = np.random.poisson(l * dt, [size, ti])
                drift = self.r - l * np.exp(self.a + self.b ** 2 / 2)

                for i in range(ti):
                    simulated_paths[:, i + 1] = simulated_paths[:, i] * np.exp((drift - self.sigma ** 2 / 2) * dt +
                                                                               self.sigma * np.sqrt(dt) * Z_1[:, i] + self.a * Poisson[:, i]
                                                                               + np.sqrt(self.b ** 2) * np.sqrt(Poisson[:, i]) * Z_2[:, i])

                return pd.DataFrame(simulated_paths)
            else:
                raise Exception(
                    "Jump diffusion model has not been calibrated yet. Calibrate it with calibrate_JD method first")
        else:
            raise Exception('''Chosen mode is not implemented. Choose mode from: "GBM", "Bootstrap", 'RT_GBM', "JD" ''')




import numpy as np
import warnings
import pandas as pd
from tqdm import tqdm


# --------------- One dimensional payoffs ----------------
def payoff_creator_1d(paths, payoff, path_dependent=False, pd_type='asian', barrier=False, barrier_level=0,
                      barrier_type='up-and-out', *args, **kwargs):
    """
    Creates various payoffs for one dimensional derivatives, depending on specified type of option and payoff function.\n
    By choosing 'path_dependent=True' you can create path dependent payoffs - that are functions of all past asset values.\n

    Parameter 'pd_type' is responsible for kind of path-dependency, possible choices:
        - 'asian', mean of asset price,
        - 'asian_geom', geometric mean,
        - 'asian_strike', strike is a mean of underlying,
        - 'asian_strike_geom', strike is a (geometric) mean,
        - 'lookback_min', min value,
        - 'lookback_max': max value,

    By setting 'barrier=True' you activate the barrier, you then need to specify it's level by setting parameter 'barrier_level'.
    Aditionally, you need to specify the type with setting 'barrier_type' to either of:
        - 'up-and-out',
        - 'down-and-out',
        - 'up-and-in',
        - 'down-and-in'.

    You might specify any additional parameters that should be passed to 'payoff_func' with *args, **kwargs. \n

    Parameters:
    -----------
    paths : array
        Generated trajectories for which we want to create a payoff in each timestep, each row of n array is a new trajectory
    payoff_func : function
        Function we want to use as payoff, taking paths as an argument and possibly other arguments
    path_dependent : bool
        If option is path dependent, it means that payoff depends on the whole (or some part) of the underlying price process
    pd_type : str
        What kind of path dependency
    barrier : bool
        If option has a barrier
    barrier_level : float
        Level of a barrier
    barrier_type : str
        Type of barrier

    Returns:
    -------
    payoffs : numpy.ndarray
        Payoffs for an option type
    """

    if type(paths) != np.ndarray:
        paths = paths.to_numpy()

    if barrier:
        if barrier_level == 0:
            raise Exception('Please specify barrier level')

        if barrier_type[:2] == 'up':
            arr = paths >= barrier_level
        elif barrier_type[:4] == 'down':
            arr = paths <= barrier_level

        # if there is no true value(path wont touch the barrier) argmax will return first false index -->0
        # as we cannot start being out of barrier I will take zeros as paths when we havent touched the barrier

        indexes = np.argmax(arr, axis=1)
        affected = np.where(indexes != 0)

    if path_dependent:
        if pd_type == 'asian':
            paths = np.cumsum(paths, axis=1) / np.arange(1, len(paths[0]) + 1)
        elif pd_type == 'asian_geom':
            paths = np.cumprod(paths, axis=1) ** (1 / np.arange(1, len(paths[0]) + 1))
        elif pd_type == 'asian_strike':
            K = np.cumsum(paths, axis=1) / np.arange(1, len(paths[0]) + 1)
        elif pd_type == 'asian_strike_geom':
            K = np.cumprod(paths, axis=1) ** (1 / np.arange(1, len(paths[0]) + 1))
        # look-back options - min/max of path
        elif pd_type == 'lookback_max':
            paths = np.maximum.accumulate(paths, axis=1)
        elif pd_type == 'lookback_min':
            paths = np.minimum.accumulate(paths, axis=1)

    payoffs = payoff(paths, *args, **kwargs)

    if barrier:
        if barrier_type[-3:] == 'out':
            for i in affected[0]:
                # print(i)
                # print(indexes[i])
                payoffs[i, indexes[i]:] = 0
            # payoffs[affected[0], indexes[affected[0]]:] = 0
            # print(affected[0][0])

        elif barrier_type[-2:] == 'in':
            payoffs_c = np.zeros(payoffs.shape)
            for i in affected[0]:
                # print(i)
                # print(indexes[i])
                payoffs_c[i, indexes[i]:] = payoffs[i, indexes[i]:]
            # payoffs[affected, :indexes[affected]] = 0
            # payoffs[-affected, :] = 0
            payoffs = payoffs_c

    return payoffs

def Call_Payoff(traj, K):
    '''
    Creates classical call payoff.

    :param traj: array of underlying values
    :type traj: numpy.ndarray
    :param K: Strike price
    :type K: float
    :return: Payoff of call option.
    :rtype: numpy.ndarray
    '''
    return np.maximum(traj-K,0)

def Put_Payoff(traj, K):
    '''
    Creates classical put payoff.

    :param traj: array of underlying values
    :type traj: numpy.ndarray
    :param K: Strike price
    :type K: float
    :return: Payoff of call option.
    :rtype: numpy.ndarray
    '''
    return np.maximum(K-traj,0)


def square_call(traj, K):
    """
    max(S^2-K, 0)

    :param traj: array of underlying values
    :type traj: numpy.ndarray
    :param K: Strike price
    :type K: float
    :return: Payoff
    :rtype: numpy.ndarray
    """
    return np.maximum(traj**2-K,0)


def straddle_payoff(traj, strike):
    """
    Calculate the payoff of a long straddle options strategy.

    The long straddle strategy involves buying a call option and a put option with the same strike price and expiration date.
    This strategy benefits from significant price movement in either direction. The call option profits when the spot price
    is above the strike price, while the put option profits when the spot price is below the strike price. The total payoff
    of the long straddle strategy is the combined payoff of the call and put options.

    Args:
        traj (numpy.ndarray): Underlying values.
        strike (float): The strike price of the call and put options.

    Returns:
        numpy.ndarray: Payoffs of the long straddle strategy.

    """
    call_payoff = Call_Payoff(traj, strike)
    put_payoff = Put_Payoff(traj, strike)
    total_payoff = call_payoff + put_payoff
    return total_payoff



def bear_put_payoff(traj, strike_long, strike_short):
    """
    Calculate the payoff of a bear put options strategy.

    The bear put strategy involves buying a put option with a higher strike price (long put) and simultaneously selling a put option with
    a lower strike price (short put). This strategy profits from a decrease in the price of the underlying asset.

    Args:
        traj (numpy.ndarray): Underlying asset.
        strike_long (float): The strike price of the long put option.
        strike_short (float): The strike price of the short put option.

    Returns:
        numpy.ndarray: The payoff of the bear put strategy.

    """
    if strike_long < strike_short:
        warnings.warn('Bear strategy assumes, that long strike is greater than short strike.')

    long_put = Put_Payoff(traj, strike_long)
    short_put = Put_Payoff(traj, strike_short)
    payoff = long_put - short_put
    return payoff


def bull_call_payoff(traj, strike_long, strike_short):
    """
    Calculate the payoff of a bull call options strategy.

    The bull call strategy involves buying a call option with a lower strike price (long call)
    and simultaneously selling a call option with a higher strike price (short call).
    This strategy profits from an increase in the spot price of the underlying asset.

    Args:
        traj (numpy.ndarray): Underlying asset.
        strike_long (float): The strike price of the long put option.
        strike_short (float): The strike price of the short put option.

    Returns:
        numpy.ndarray: The payoff of the bull call strategy.

    """
    if strike_long > strike_short:
        warnings.warn('Bull strategy assumes, that long strike is smaller than short strike')

    long_call = Call_Payoff(traj, strike_long)
    short_call = Call_Payoff(traj, strike_short)
    payoff = long_call - short_call
    return payoff



# ---------------- Two dimensional ------------------------

def spread_payoff(traj_list):
    """
    Calculate the spread payoff for two assets.

    The spread payoff is maximum of the difference between the prices of the two assets and 0.

    Args:
        traj (list): List of trajectories

    Returns:
        numpy.ndarray: The spread payoff.

    """
    if len(traj_list) != 2:
        raise Exception("This payoff is created for two dimensional options!")
    return np.maximum(traj_list[0]-traj_list[1], 0)


def ratio_spread_payoff(traj_list, asset1_weight=1,asset2_weight=1):
    """
    Calculate the ratio spread payoff for two assets.

    The ratio spread payoff is similar to the spread payoff but involves a different number of positions in each asset.

    Args:
        traj_list (list): Underlying assets.
        asset1_weight (float): Number of asset1 to include.
        asset2_weight (float): Number of asset2 to include.

    Returns:
        numpy.ndarray: The ratio spread payoff.

    """
    if len(traj_list) != 2:
        raise Exception("This payoff is created for two dimensional options!")
    return np.maximum(asset1_weight*traj_list[0] - asset2_weight* traj_list[1], 0)


# ------------ Any number of assets ------------
def multiple_call_payoff(traj_list, K):
    """
    Calculate the call payoff for multiple assets.

    The multiple call payoff is calculated as the maximum of the call payoffs for each asset,
    you can specify different strike for every asset (pass as list) or same strike for all.

    Args:
        traj_list (list): Underlyings trajectories.
        K (list or float): The strike price or list of strike prices for each asset.

    Returns:
        numpy.ndarray: Payoffs

    """
    n_assets = len(traj_list)
    if isinstance(K, (int, float)):
        K = [K] * n_assets
    if len(K) != n_assets:
        raise Exception("Please specify one strike for each asset!")
    payoff = np.zeros(traj_list[0].shape)
    for a in range(n_assets):
        payoff = np.maximum(payoff, Call_Payoff(traj_list[a] , K[a]))
    return payoff

def multiple_put_payoff(traj_list, K):
    """
    Calculate the put payoff for multiple assets.

    The multiple put payoff is calculated as the maximum of the put payoffs for each asset,
    you can specify different strike for every asset (pass as list) or same strike for all.

    Args:
        trajs (list): Underlyings trajectories.
        K (list or float): The strike price or list of strike prices for each asset.

    Returns:
        numpy.ndarray: Payoffs

    """
    n_assets = len(traj_list)
    if isinstance(K, (int, float)):
        K  = [K] * n_assets
    if len(K) != n_assets:
        raise Exception("Please specify one strike for each asset!")
    payoff = np.zeros(traj_list[0].shape)
    for a in range(n_assets):
        payoff = np.maximum(payoff, Put_Payoff(traj_list[a], K[a]))
    return payoff


def multiple_mixed_payoff(traj_list, K, payoff_type, which='max', barrier = False, barrier_level=0, barrier_type='up-and-out'):
    """
    Calculate the payoff for multidimensional mixed payoff option.

    The multiple mixed payoff is calculated as the maximum (minimum) of the call or put payoffs for each asset,
    you can specify different strike for every asset (pass as list) or same strike for all.

    You have to specify the type of option for each asset by passing a list of strings
    'call', 'put' to the 'payoff_type' argument.

    By setting parameter which='min' you will take the minimum of all payoffs.
    The default is which='max', that is maximum of all payoffs.

    Args:
        trajs (list): Underlyings trajectories.
        K (list or float): The strike price or list of strike prices for each asset.
        payoff_type (list): The list of strings 'call', 'put' to specify the type for each asset.
        which (string): Default 'max' - we choose maximal payoff of all, 'min' - minimal payoff.

    Returns:
        numpy.ndarray: Payoffs

    """
    n_assets = len(traj_list)
    if isinstance(K, (int, float)):
        K = [K] * n_assets
    if len(payoff_type) != n_assets:
        raise Exception("Please specify one payoff type for each asset!")
    if len(K) != n_assets:
        raise Exception("Please specify one strike for each asset!")
    if which not in ['max', 'min']:
        raise Exception("Only possible values for which argument are 'max' or 'min'!")

    payoff = np.zeros(traj_list[0].shape)
    for a in range(n_assets):
        if payoff_type[a] == 'put':
            p = Put_Payoff(traj_list[a], K[a])
        elif payoff_type[a] == 'call':
            p = Call_Payoff(traj_list[a], K[a])
        else:
            raise Exception("The only supported values for payoff type are 'call' and 'put'!")

        if which == 'max':
            payoff = np.maximum(payoff, p)
        elif which == 'min':
            if a == 0:
                payoff = p
            else:
                payoff = np.minimum(payoff, p)

    return payoff

multiple_call_payoff.__str__ = lambda: 'multiple call payoff'

def Double_Call_Payoff(sims_list, K1, K2):
    return (np.maximum(sims_list[0] - K1, 0) + np.maximum(sims_list[1] - K2, 0))

#def double_max_put(sims_list, K):
#    return np.maximum(np.maximum(K - sims_list[0], K - sims_list[1]), 0)






from setuptools import setup

from codecs import open
from os import path

# The directory containing this file
HERE = path.abspath(path.dirname('setup.py'))

# Get the long description on pypi from the README file
with open(path.join(HERE, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# This call to setup() does all the work
setup(
    name="american_options",
    version="0.1.7",
    description="Updated plots",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://american_options.readthedocs.io/",
    author="Przemysław Adamski, Katarzyna Hasal, Kacper Toczek",
    author_email="kat.hasal99@gmail.com",
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent"
    ],
    packages=["american_options"],
    include_package_data=True,
    install_requires=["numpy",
                      "scipy",
                      "chaospy",
                      "pandas",
                      "matplotlib",
                      "sklearn",
                      "tqdm"]
)